unit ICeTEe;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	ICeTEe
	------
	This file implements the bulk of the ICeTEe extension.  The basic premise here is
	that we have an INIT that runs at startup time and patches _TEClick.  When the
	user clicks in a TextEdit field, the application program calls _TEClick.  Our
	patch latches the value of the command key, calls through to the old _TEClick
	routine and then, if the command key was down, does its stuff.
	
	All the rest of the code is basically support for this primary function.
	
	There is also a Gestalt selector so that people can:
	
	a) determine whether ICeTEe is installed
	b) write a UI to edit the exclusion list without requiring a restart for it to
		take effect
*)

interface

{$main}
	procedure Main;

implementation

	uses
		Processes,
		Notification,
		Traps,
		TextUtils,
		GestaltEqu,
		TextEdit,
		Errors,
		Resources,
		LowMem,
		
		PascalA4,

		ShowInitIcon, 

		InternetConfig;

	const
		kICeTEeCreator = 'ICTE';

	(* ***** Resources ***** *)
	
	const
		(* INIT *)
		rInitCode = 128;
		
		(* EXCL *)
		rExclusions = 128;

		(* ICN# *)
		rICTEIcon = 128;
		rFailedIcon = 129;

		(* STR# *)
		rErrorStrings = 128;
		strMiscErr = 1;
		strNoCMErr = 2;
		strNoICErr = 3;
		strInsufficientICErr = 4;
		strNoMemoryErr = 5;
		strCantFindHelperErr = 6;
		strNoHelperErr = 7;
		strNoURLErr = 8;
		strCantHackIt = 9;

	(* ***** Extra Error Codes ***** *)
	
	const
		noCMErr = -6660;

	(* ***** Standard Utility Routines ***** *)

	type
		IntegerHandle = ^IntegerPtr;
		(* InternetPtr is defined in Types.p, but not IntegerHandle.  *sigh* *)

	function GetIndStrList (strListH: Handle; index: integer): Str255;
		(* This function returns the index'th string in the string list. If index
			is out of bounds, it returns the empty string.
			
			Stolen directly from ICStringLists.p in the IC Application Source
			Kit.
		*)
		var
			result: Str255;
			count : integer;
			stringListIndex: integer;
			position: longint;
	begin
		count := IntegerHandle(strListH)^^;
		if (1 <= index) and (index <= count) then begin
			position := sizeof(integer);
			for stringListIndex := 1 to index - 1 do begin
				position := position + band(Ptr(ord(strListH^) + position)^, $FF) + 1;
			end; (* for *)
			BlockMoveData(Ptr(ord(strListH^) + position), 
									@result, 
									band(Ptr(ord(strListH^) + position)^, $FF) + 1);
		end else begin
			result := '';
		end; (* if *)
		GetIndStrList := result;
	end; (* GetIndStrList *)

	function DecStr (aNumber: longint): Str255;
		(* This is a simpl wrappers around the toolbox NumToString routine.

			Stolen directly from ICMiscSubs.p in the IC Application Source
			Kit.
		*)
		var
			result : Str255;
	begin
		NumToString(aNumber, result);
		DecStr := result;
	end; (* DecStr *)

	procedure BlockFill (baseAddr: univ Ptr; size: longint; value: integer);
		(* Fills a block of memory with the value.  The memory is filled
			as bytes, ie the high byte of value is ignored.

			Stolen directly from ICMiscSubs.p in the IC Application Source
			Kit.
		*)
	begin
		while (size > 0) do begin
			baseAddr^ := value;
			inc(longint(baseAddr));
			size := size - 1;
		end; (* while *)
	end; (* BlockFill *)

	function HaveComponentManager: Boolean;
		(* Returns true if the Component Manager is available on this machine. *)
		var
			response: longint;
	begin
		HaveComponentManager := (Gestalt(gestaltComponentMgr, response) = noErr);
	end; (* HaveComponentManager *)

	(* ***** Core Data Structures ***** *)
	
	(* The exclusion array is used to hold an array of OSTypes that are
		the application creators for which ICeTEe should not function.
		Some applications use TextEdit but don't require ICeTEe support,
		mainly because they implement it internally.  This exclusion
		array prevents ICeTEe from causing the URL to be launched
		twice.
	*)
	
	type
		ExclusionArray = array[1..1000] of OSType;
		ExclusionArrayPtr = ^ExclusionArray;
		ExclusionArrayHandle = ^ExclusionArrayPtr;

	(* The ICeTEe globals are held in one global record.  This is
		mostly a legacy of the original Think Pascal implementation,
		in which ICeTEe didn't use global variables but instead
		allocated its globals in the system heap and simply stored
		a pointer to them.  Because I'm now using CodeWarrior
		(which requires you to set up A4 even if you're not using globals)
		I could use proper global variables, but I've chosen not to.
		The reason: ICeTEe publishes the address of its globals through
		a Gestalt selector.  These globals have a defined layout.  If
		I moved the globals into true global variables, there's a possibility
		that the layout might change.  So the globals are still all in
		a record, one instance of which is allocated globally.
	*)
	
	type
		TEClickProcPtr = procedure (pt:Point; fExtend:Boolean; h:TEHandle);
	
	type
		ICeTEeGlobals =
			record
				signature: OSType;							(* Set to kICeTEeCreator. *)
				version: NumVersion;						(* Version of ICeTEe. *)
				exclusionList: ExclusionArrayHandle;	(* List of excluded application creators. *)
				errorStringList : Handle;									(* STR# that lists the possible error strings. *)
				oldTEClickProc: TEClickProcPtr;		(* Old trap address. *)
			end;
		ICeTEeGlobalsPtr = ^ICeTEeGlobals;

	(* The single instance of ICeTEeGlobals is allocated globally. *)
	
	var
		gGlobals : ICeTEeGlobals;
		
	function GetMyGlobals: ICeTEeGlobalsPtr;
		(* Returns the address of the ICeTEeGlobals.  Used to be a horrible
			Think Pascal hack, now it just returns the address of the global
			variable gGlobals.
		*)
	begin
		GetMyGlobals := @gGlobals;
	end; (* GetMyGlobals *)

	(* ***** Core Implementation ***** *)
	
	function CurrentProcessExcluded: Boolean;
		(* Returns true if the current process has a signature that is in the exclusion
			list.
		*)			
		var
			currentPSN : ProcessSerialNumber;
			currentProcessInfo : ProcessInfoRec;
			exclusionList : ExclusionArrayHandle;
			exclusionIndex : integer;
	begin
		(* Set up the call to GetProcessInformation to get info about the
			current process.
		*)
		currentPSN.highLongOfPSN := 0;
		currentPSN.lowLongOfPSN := kCurrentProcess;
		currentProcessInfo.processInfoLength := sizeof(ProcessInfoRec);
		currentProcessInfo.processName := nil;
		currentProcessInfo.processAppSpec := nil;
		if GetProcessInformation(currentPSN, currentProcessInfo) = noErr then begin

			(* Search the exclusion list, looking for the current process' creator.
				If we find it, this process is excluded and we return true.
			*)
			exclusionList := GetMyGlobals^.exclusionList;
			CurrentProcessExcluded := false;
			for exclusionIndex := 1 to GetHandleSize(Handle(exclusionList)) div 4 do begin
				if exclusionList^^[exclusionIndex] = currentProcessInfo.processSignature then begin
					CurrentProcessExcluded := true;
					leave;
				end; (* if *)
			end; (* for *)

		end else begin
			CurrentProcessExcluded := true;
		end; (* if *)
	end; (* CurrentProcessExcluded *)

	function DoCommandClick (teh: TEHandle; selStart, selEnd: longint): ICError;
		(* This routine represents the core implementation of command-clicking.
			It is called by ICeTEeClick when it has officially determined that this
			is a command-click that we're going to act on. This routine
			connects up with Internet Config and calls ICLaunchURL on the
			text.
		*)
		var
			inst: ComponentInstance;
			err: ICError;
			err2: ICError;
			text: Handle;
			s: signedByte;
			i: integer;
			junklong: longint;
			hint, at: Str31;
			urlh: Handle;
			tmpSelStart, tmpSelEnd: longint;
	begin
		(* Open a connection to the IC component. *)
		if HaveComponentManager then begin
			err := ICCStart(inst, kICeTEeCreator);
		end else begin
			err := noCMErr;
		end; (* if *)
		
		if err = noErr then begin
		
			(* Specify the default config file. *)
			err := ICCFindConfigFile(inst, 0, nil);
			if err = noErr then begin

				(* Set up for parsing the URL. *)
				text := Handle(TEGetText(teh));
				s := HGetState(text);
				HLock(text);
				urlh := NewHandle(0);
				hint := 'mailto';
				tmpSelStart := selStart;
				tmpSelEnd := selEnd;
				
				(* Call ICParseURL to determine whether there is a URL there to begin with.
					If there is, look in the parsed out URL to see whether it contains an '@'
					sign.  If it does, supply a "mailto" hint to ICLaunchURL, otherwise
					we supply no hint.
					
					This is basically a simple form of slack URl handling.  If the URL we're
					about to launch contains an "@", we give IC a "mailto" hint.  This means
					that URLs of the form <internet-config@share.com> will be launched
					as <mailto:internet-config@share.com>, without otherwise affecting the
					URL parsing scheme of IC itself.
				*)
				err := ICCParseURL(inst, hint, text^, GetHandleSize(text), tmpSelStart, tmpSelEnd, urlh);
				if err = noErr then begin
					hint := '';
					at := '@';
					if Munger(urlh, 0, @at[1], length(at), nil, 0) >= 0 then begin
						hint := 'mailto';
					end; (* if *)
					err := ICCLaunchURL(inst, hint, text^, GetHandleSize(text), selStart, selEnd);
				end; (* if *)
				DisposeHandle(urlh);

				(* Now flash the launched URL by selecting it and repeatedly activating/
					deactivating the TEHandle.  Note that we leave the URL selected, even
					if we're returning an error.  This means that, if there is an error,
					the erroneous URL is highlighted.
				*)
				TESetSelect(selStart, selEnd, teh);
				if err = noErr then begin
					for i := 1 to LMGetMenuFlash do begin
						Delay(5, junklong);
						TEDeactivate(teh);
						Delay(5, junklong);
						TEActivate(teh);
					end; (* for *)
				end; (* if *)
				HSetState(text, s);
			end; (* if *)
			
			(* Clean up. *)
			err2 := ICCStop(inst);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		
		DoCommandClick := err;
	end; (* DoCommandClick *)

	procedure MyNMResponseProc (nm: NMRecPtr);
		(* When ICeTEe fails to launch a URL, it posts a notification saying
			why.  This routine is the response proc for that notification.
			It's mission in life is to dispose of the NMRec in a clean fashion.
		*)
		var
			oldA4 : longint;
			oldZone : THz;
			junk : OSErr;
	begin
		(* Set up our environment. *)
		oldA4 := SetUpA4;
		oldZone := GetZone;
		SetZone(SystemZone);
		
		(* Remove the notification. *)
		junk := NMRemove(nm);
		
		(* Dispose of the notification text string and the notification itself. *)
		if nm^.nmStr <> nil then begin
			DisposePtr(Ptr(nm^.nmStr));
		end; (* if *)
		DisposePtr(Ptr(nm));

		(* Restore caller's environment. *)
		SetZone(oldZone);
		oldA4 := SetA4(oldA4);
	end; (* MyNMResponseProc *)

	procedure PostErrorNotification(errNum : ICError);
		var
			errorStringListIndex : integer;
			errorString : Str255;
			offsetToErrorNumberInErrorString : integer;
			notification: NMRecPtr;
			errorStringPtr : Ptr;
			junk : OSErr;
	begin
		(* Convert the error number into an index into the error string list. *)
		case errNum of
			badComponentInstance :
				errorStringListIndex := strNoICErr;
			noCMErr :
				errorStringListIndex := strNoCMErr;
			badComponentSelector :
				errorStringListIndex := strInsufficientICErr;
			memFullErr : 
				errorStringListIndex := strNoMemoryErr;
			afpItemNotFound :
				errorStringListIndex := strCantFindHelperErr;
			icPrefNotFoundErr :
				errorStringListIndex := strNoHelperErr;
			icNoURLErr :
				errorStringListIndex := strNoURLErr;
			noPortErr :
				errorStringListIndex := strCantHackIt;
			otherwise
				errorStringListIndex := strMiscErr;
		end; (* case *)

		(* Now get the corresponding message from the list of errors we keep globally. *)
		errorString := GetIndStrList(GetMyGlobals^.errorStringList, errorStringListIndex);

		(* If we succeed, let's substitute the error number for ^0 (if there is one). *)
		if errorString <> '' then begin
			offsetToErrorNumberInErrorString := Pos('^0', errorString);
			if offsetToErrorNumberInErrorString <> 0 then begin
				Delete(errorString, offsetToErrorNumberInErrorString, length('^0'));
				Insert(DecStr(errNum), errorString, offsetToErrorNumberInErrorString);
			end; (* if *)
			
			(* Now create the notification record, putting the error string in
				a separate pointer block.
			*)
			errorStringPtr := NewPtrSys(length(errorString) + 1);
			if errorStringPtr <> nil then begin
				BlockMoveData(@errorString[0], errorStringPtr, length(errorString) + 1);
			end; (* if *)
			notification := NMRecPtr(NewPtrSysClear(sizeof(NMRec)));
			if notification <> nil then begin

				(* Post the notification. *)
				notification^.qType := ord(nmType);
				notification^.nmMark := 0;
				notification^.nmIcon := nil;
				notification^.nmSound := nil;
				notification^.nmStr := StringPtr(errorStringPtr);
				notification^.nmResp := @MyNMResponseProc;
				junk := NMInstall(notification);

			end else begin
				SysBeep(10);
			end; (* if *)
		end; (* if *)
	end; (* PostErrorNotification *)
	
	procedure ICeTEeClick (teh: TEHandle; oldSelStart, oldSelEnd: integer);
		(* This is called by core _TEClick patch routine when we know it's
			a command-click.  We establish our
			environment, check for exclusions, massage the text offsets,
			do the command-click, and then post the error notification if necessary.
		*)
		var
			err: ICError;
			oldZone : THz;
	begin
		(* Save the current heap zone, and switch to the system zone for all our stuff. *)
		oldZone := GetZone;
		SetZone(SystemZone);

		(* If we're not running in an excluded process. *)
		if not CurrentProcessExcluded then begin
			
			(* Adjust oldSelStart and oldSelEnd in some way I currently don't understand, but
				I remember it doing something useful!
			*)
			if not (		(oldSelStart <= teh^^.selStart) and 
							(teh^^.selStart <= oldSelEnd) and 
							(oldSelStart <= teh^^.selEnd) and 
							(teh^^.selEnd <= oldSelEnd)) then begin
				oldSelStart := teh^^.selStart;
				oldSelEnd := teh^^.selEnd;
			end; (* if *)
			
			(* Do the command-click, and post any resulting error. *)
			err := DoCommandClick(teh, oldSelStart, oldSelEnd);
			if err <> noErr then begin
				PostErrorNotification(err);
			end; (* if *)
		end; (* if *)
		
		(* Clean up. *)
		SetZone(oldZone);
	end; (* ICeTEeClick *)

	procedure PascalTEClickPatch (pt: Point; fExtend: Boolean; teh: TEHandle);
		(* This is the core patch on _TEClick.  It remembers samples the command
			key and remembers the current selection,
			calls through to the old _TEClick implementation, and then, if it
			was a command-click, calls ICeTEeClick.
		*)
		var
			oldA4 : longint;
			oldSelStart, oldSelEnd: integer;
			km: KeyMap;
			commandKeyDown: Boolean;
	begin
		oldA4 := SetUpA4;
		
		(* Remember the selection and the command key state prior to the click. *)
		oldSelStart := teh^^.selStart;
		oldSelEnd := teh^^.selEnd;
		GetKeys(km);
		commandKeyDown := km[55];
		
		(* Call through to the old _TEClick implementation. *)
		GetMyGlobals^.oldTEClickProc(pt, fExtend, teh);
		
		(* If it's a command-click, go do stuff. *)
		if commandKeyDown and (GetHandleSize(Handle(TEGetText(teh))) > 0) then begin
			ICeTEeClick(teh, oldSelStart, oldSelEnd);
		end; (* if *)
		
		oldA4 := SetA4(oldA4);
	end; (* PascalTEClickPatch *)

	procedure AsmTEClickPatch(pt: Point; fExtend: Boolean; teh: TEHandle);
		(* This is a simple wrapper around PascalTEClickPatch.  I use it because
			I want to ensure that all the registers are saved within my patch,
			and the alternative way of doing this was ridiculously ugly.
		*)
		asm;
	begin
					link			a6,#0								(* create a stack frame *)
					movem.l	a0-a5/d0-d7,-(sp)			(* save *all* the registers *)

					move.l		14(a6),-(sp)						(* call the PascalTEClickPatch *)
					move.w		12(a6),-(sp)
					move.l		8(a6),-(sp)
					jsr			PascalTEClickPatch

					movem.l	(sp)+,a0-a5/d0-d7			(* recover registers and tear down frame *)
					move.l		4(a6),a0
					unlk			a6
					lea			14(a7),a7							(* remove params and return addr *)
					jmp			(a0)
	end; (* AsmTEClickPatch *)

	function ICeTEeGestalt (selector: OSType; var response: longint): OSErr;
		(* This is ICeTEe's Gestalt responder function.  It simply returns
			the address of the IC globals.
		*)
		var
			oldA4 : longint;
	begin
		{$unused selector}
		oldA4 := SetUpA4;
		response := longint(GetMyGlobals);
		ICeTEeGestalt := noErr;
		oldA4 := SetA4(oldA4);
	end; (* ICeTEeGestalt *)

	procedure Main;
		(* This is the INIT's mainline.  It is responsible for:
		
			o showing the icon
			o checking the environment
			o initialising the ICeTEe globals
			o installing the patches
			o registering the Gestalt
			o showing the success or failure icons
			
			It's kinda long but not very complicated.
		*)
		
		var
			oldA4 : longint;
			err: OSErr;
			err2: OSErr;
			response: longint;
			ozone: THz;
			exclusionList: Handle;
			errorStringList: Handle;
			vers: VersRecHndl;
			oldTEClickProc : longint;
	begin
(*		Debugger; *)

		(* Initialise our A4 world and remember it for the callbacks. *)
		oldA4 := SetCurrentA4;
		RememberA4;

		(* Detach the INIT resource, so we don't get cleaned up when our
			resource file is closed.  Note that the INIT code resource is locked in
			the system heap because of the resource atttributes.
		*)
		DetachResource(GetResource('INIT', rInitCode));

		(* Show the ICeTEe icon to say that we've started loading. *)
		ShowInitIcon(rICTEIcon, false);
		
		(* Check our environment.  ICeTEe requires System 7. *)
		err := noErr;
		if (Gestalt(gestaltSystemVersion, response) <> noErr) | (response < $700) then begin
			err := unimpErr;
		end; (* if *)
		
		(* Create and initialise the globals. *)
		if err = noErr then begin
			ozone := GetZone;
			SetZone(SystemZone);

			(* Globals are initialised to nil by the CodeWarrior runtime, so 
				we don't need to explicitly clear them ourselves.
			BlockFill(@gGlobals, sizeof(gGlobals), 0);
			*)
			
			(* Initialise the globals. *)
			gGlobals.signature := kICeTEeCreator;

			vers := VersRecHndl(Get1Resource('vers', 1));
			if vers <> nil then begin
				gGlobals.version := vers^^.numericVersion;
			end; (* if *)

			exclusionList := Get1Resource('EXCL', rExclusions);
			err := HandToHand(exclusionList);
			gGlobals.exclusionList := ExclusionArrayHandle(exclusionList);

			errorStringList := Get1Resource('STR#', rErrorStrings);
			err2 := HandToHand(errorStringList);
			gGlobals.errorStringList := errorStringList;

			(* Clean up. *)
			if err = noErr then begin
				err := err2;
			end; (* if *)
			SetZone(ozone);
		end; (* if *)

		(* Register our Gestalt selector.  Note that we don't try to replace an existing
			selector.  If the selector already exists, it's likely the user has a second
			copy of ICeTEe installed, so we just bale out with an error.
		*)
		if err = noErr then begin
			err := NewGestalt(kICeTEeCreator, @ICeTEeGestalt);
		end; (* if *)
		
		(* Now we install our patch.  We do this last so that we don't have any
			real effect on the operation of the system unless all our initialisation
			was successful.
		*)
		if err = noErr then begin
			oldTEClickProc := longint(NGetTrapAddress(_TEClick, ToolTrap));
			gGlobals.oldTEClickProc := TEClickProcPtr(oldTEClickProc);
			NSetTrapAddress(UniversalProcPtr(@AsmTEClickPatch), _TEClick, ToolTrap);
		end; (* if *)

		(* If we got an error then we bleed memory all over the place.  This is not *)
		(* an accident.  How many copies of the init can you reasonably fail to install??? *)

		(* Now show the success or failure icon. *)
		if err = noErr then begin
			ShowInitIcon(rICTEIcon, true);
		end else begin
			ShowInitIcon(rFailedIcon, true);
		end; (* if *)

		(* Restore environment. *)
		oldA4 := SetA4(oldA4);
	end; (* Main *)

end. (* ICeTEe *)
